local _,stash = ...
local C = stash.constants

--[[
    This file is for tracking itemIDs as a player acquires/loses them and where they are.

    Expected behaviors:
    - For various reasons, this addon only tracks itemIDs and not any customization done on
      items. For instance, 6/6 Korthian armor will be their base armor. Plate Boots of the Eagle
      will be just Plate Boots without the suffix. Enchants, gems, etc. will not be attached to
      any item.
    - On Classic realms, sending mail to another character still takes an hour. If you check
      mail on the other character before it's arrived, it will remove the incoming item
      from the other character's mail stash (until you check mail when it's arrived). This addon
      does not attempt to track time it takes for mail to arrive.

]]

StashSaved = {} -- everything is saved in StashSaved savedvar, indexed by realm and then character name

local dirtyBags = CopyTable(C.INVENTORY_BAGS) -- bag slots (0-4) that change in BAG_UPDATE (login as all dirty)
local dirtyBank = CopyTable(C.BANK_BAGS) -- bank bags (5-11) that change in BAG_UPDATE

local isBankOpen -- true while bank is open
local isGuildBankOpen -- true while guild bank is open

local currentGuildBankQueryTab -- guild bank tab being queried

local callbacks = {} -- indexed by "item" or "money", functions to call when items or money has changed

local pendingSendMail = {} -- ordered list with first row as {realm,recipient} and rest of rows as {itemID,count}

-- self.tracking is an independent frame solely for item tracking events and methods
stash.tracking = CreateFrame("Frame")
stash.tracking:SetScript("OnEvent",function(self,event,...)
    if self[event] then
        self[event](self,...)
    end
    if callbacks.item then
        callbacks.item()
    end
    stash.info:ResetLocationsCache()
end)

-- housekeeping done once on login
function stash.tracking:Setup()
    self:RegisterEvent("BAG_UPDATE")
    self:RegisterEvent("BAG_UPDATE_DELAYED")
    self:RegisterEvent("PLAYER_EQUIPMENT_CHANGED")
    self:RegisterEvent("ITEM_UNLOCKED")
    self:RegisterEvent("PLAYER_MONEY")
    self:RegisterEvent("BANKFRAME_OPENED")
    self:RegisterEvent("BANKFRAME_CLOSED")
    self:RegisterEvent("PLAYERBANKSLOTS_CHANGED")

    self:RegisterEvent("MAIL_CLOSED")
    self:RegisterEvent("MAIL_SEND_SUCCESS")
    self:RegisterEvent("MAIL_INBOX_UPDATE")

    if C.IS_MAINLINE then
        self:RegisterEvent("PLAYERREAGENTBANKSLOTS_CHANGED")
        self:RegisterEvent("CURRENCY_DISPLAY_UPDATE")
        self:RegisterEvent("VOID_TRANSFER_DONE")
    end
    if C.IS_BCC or C.IS_MAINLINE then
        self:RegisterEvent("GUILDBANKFRAME_OPENED")
        self:RegisterEvent("GUILDBANKFRAME_CLOSED")
        self:RegisterEvent("GUILDBANKBAGSLOTS_CHANGED")
    end

    -- on login, take an inventory of accessible items and money on character
    self:BAG_UPDATE_DELAYED()
    for slot=1,23 do -- and update worn stuff
        self:PLAYER_EQUIPMENT_CHANGED(slot)
    end
    if C.IS_MAINLINE then
        if not C.IS_MODERN_CLIENT then
            self:UpdateContainer(stash.current.reagents,-3) -- reagents are always available in Shadowland clients
        end
        self:UpdateAllCurrency() -- updates player's currency
        self:VOID_TRANSFER_DONE()
    end
    self:PLAYER_MONEY()

    hooksecurefunc("SendMail",stash.tracking.SendMailHook) -- watch for mail being sent
end

-- when either "item" or "money" changes in tracking, call the given function. multiple
-- modules can call this and it will run their functions for the tracking type in the order
-- they were "registered"
function stash.tracking:SetCallback(callbackType,func)
    if type(func)~="function" or (callbackType~="item" and callbackType~="money") then
        return
    end
    if callbacks[callbackType] then
        local old = callbacks[callbackType]
        callbacks[callbackType] = function()
            old()
            func()
        end
    else
        callbacks[callbackType] = func
    end
end

--[[ events ]]

function stash.tracking:BAG_UPDATE(bag)
    if C.INVENTORY_BAGS[bag] then -- for worn bags that are changing
        dirtyBags[bag] = true
    elseif C.BANK_BAGS[bag] then -- otherwise it's a bank bag that needs to wait until bank open
        dirtyBank[bag] = true
    end
end

-- fired after a set of BAG_UPDATEs; go through each bag that had an update and update their contents
function stash.tracking:BAG_UPDATE_DELAYED()
    for bag in pairs(dirtyBags) do
        stash.tracking:UpdateContainer(stash.current.bags,bag)
    end
    wipe(dirtyBags)
    if isBankOpen then
        for bag in pairs(dirtyBank) do
            stash.tracking:UpdateContainer(stash.current.bank,bag)
        end
        self:UpdateContainer(stash.current.reagents,-3)
        wipe(dirtyBank)
    end
end

-- nothing stacked can be worn (ammo in classic exists in bags)
function stash.tracking:PLAYER_EQUIPMENT_CHANGED(slot)
    stash.current.worn[slot] = GetInventoryItemID("player",slot)
end

-- worn bags are inventory slots 20-23
function stash.tracking:ITEM_UNLOCKED(slot)
    if slot and slot>=20 and slot<=23 then
        stash.current.worn[slot] = GetInventoryItemID("player",slot)
    end
end

-- when player's money changes; this value is an integer of copper amount
function stash.tracking:PLAYER_MONEY()
    stash.current.money = GetMoney()
    if callbacks.money then
        callbacks.money()
    end
end

-- there is no equivalent of BAG_UPDATE_DELAYED for reagent bank tab, just a per-slot change
function stash.tracking:PLAYERREAGENTBANKSLOTS_CHANGED(slot)
    if isBankOpen or not C.IS_MODERN_CLIENT then
        self:UpdateContainerSlot(stash.current.reagents,-3,slot)
    end
end

-- this is for the main bank area (bag slot -1) and in retail can fire if quest/tradeskill items used; but if the bank
-- is not open then we can't scan changes yet
function stash.tracking:PLAYERBANKSLOTS_CHANGED(slot)
    if isBankOpen then
        self:UpdateContainerSlot(stash.current.bank,-1,slot)
        dirtyBank[-1] = true -- kinda redundant, since whole bank scanned when opened; but being safe
    end
end

-- when currency changes, can directly update the changing currency
function stash.tracking:CURRENCY_DISPLAY_UPDATE(currencyID,quantity)
    stash.current.currency[format("currency:%d",currencyID)] = quantity
    if callbacks.item then
        callbacks.item()
    end
end

function stash.tracking:BANKFRAME_OPENED()
    isBankOpen = true
    -- it's possible that reagents or quest items in bank were consumed, rescan whole bank when it's opened
    for bag in pairs(stash.current.bank) do
        dirtyBank[bag] = true
    end
    self:UpdateContainer(stash.current.reagents,-3)
    -- after a fraction of a second, pick up bank contents
    C_Timer.After(C.BANK_OPEN_DELAY,function()
        stash.tracking:BAG_UPDATE_DELAYED()
        if callbacks.item then
            callbacks.item()
        end
    end)
end

function stash.tracking:BANKFRAME_CLOSED()
    isBankOpen = false
end

--[[ guild bank ]]

function stash.tracking:GUILDBANKFRAME_OPENED()
    if not isGuildBankOpen then -- GUILDBANKFRAME_OPENED can run twice, only doing stuff on first
        -- first send queries for tab contents (taking a small hit on initial open; doing one tab per server call wasn't much better)
        for tab=1,GetNumGuildBankTabs() do
            if select(3,GetGuildBankTabInfo(tab)) then -- only query tabs player can view
                QueryGuildBankTab(tab) -- need to query server for contents; triggers GUILDBANKSLOTS_CHANGED
            end
        end
        -- create a saved user under guild name if one isn't made yet
        local saved = StashSaved
        local guildName = GetGuildInfo("player")
        if not guildName then -- shouldn't be possible to open a guild bank if not in a guild, but just to be safe
            return
        end
        if not saved[stash.current.realm][guildName] then
            saved[stash.current.realm][guildName] = {
                realm = stash.current.realm,
                player = guildName,
                faction = stash.current.faction,
                money = 0,
                isGuildBank = true,
                guild = {}
            }
        end
        saved = saved[stash.current.realm][guildName]
        if not saved.guild then saved.guild = {} end -- in case it got malformed somehow
        saved.update = date("%x")
    end
    isGuildBankOpen = true
end

function stash.tracking:GUILDBANKFRAME_CLOSED()
    isGuildBankOpen = false
end

-- even fired when QueryGuildBankTab results returned and contents of a tab changes
function stash.tracking:GUILDBANKBAGSLOTS_CHANGED()
    if isGuildBankOpen then
        stash.timer:Start(C.GUILDBANK_OPEN_DELAY,stash.tracking.UpdateGuildBank)
    end
end

function stash.tracking:UpdateGuildBank()
    if not isGuildBankOpen then
        return
    end
    local saved = StashSaved[stash.current.realm][GetGuildInfo("player")]
    if saved then
        for tab=1,GetNumGuildBankTabs() do
            if select(3,GetGuildBankTabInfo(tab)) then -- only for visible tabs
                if not saved.guild[tab] then
                    saved.guild[tab] = {}
                end
                for slot=1,C.MAX_GUILDBANK_SLOTS_PER_TAB do
                    if GetGuildBankItemLink(tab,slot) then
                        if not stash.tracking:UpdateGuildBankSlot(saved.guild,tab,slot) then
                            stash.timer:Start(C.DATA_CACHE_DELAY,stash.tracking.UpdateGuildBank)
                        end
                    else
                        saved.guild[tab][slot] = nil
                    end
                end
            end
        end
        saved.money = GetGuildBankMoney()
    end
end

--[[ helper methods for events ]]

function stash.tracking:GetContainerNumSlots(bag)
    return C_Container.GetContainerNumSlots(bag)
end

function stash.tracking:GetContainerItemID(bag,slot)
    return C_Container.GetContainerItemID(bag,slot)
end

function stash.tracking:GetContainerItemInfo(bag,slot)
    local info = C_Container.GetContainerItemInfo(bag,slot)
    return info.iconFileID, info.stackCount, info.isLocked, info.quality, info.isReadable, info.hasLoot, info.hyperlink, info.isFiltered, info.hasNoValue, info.itemID, info.isBound
end

-- for bags, bank or reagents, updates all contents of the given bag in the given savedvar table
-- we need to get use GetContainerItemInfo to get a count, but it's not valid if an itemID isn't data cached
-- container is one of: stash.current.bags, stash.current.bank or stash.current.reagents
-- bag is 0-4 for bags, 5-11 for bank, -3 for reagents
function stash.tracking:UpdateContainer(container,bag)
    if (container==stash.current.bank or container==stash.current.reagents) and not isBankOpen then
        return -- we were about to update bank bags while bank isn't open; leave
    end
    local needsDataCache = false -- becomes true if anything isn't data cached
    if container[bag] then
        wipe(container[bag])
    end
    for slot=1,self:GetContainerNumSlots(bag) do
        if not self:UpdateContainerSlot(container,bag,slot) then
            needsDataCache = true
        end
    end
    if needsDataCache then -- if something wasn't data cached, come back shortly and try again
        C_Timer.After(C.DATA_CACHE_DELAY,function() -- using a closure here so the same container,bag is retried
            stash.tracking:UpdateContainer(container,bag)
            if callbacks.item then
                callbacks.item()
            end
        end)
    end
end

-- updates the bag,slot of the given container and returns true if it was successfully cached; false if a data cache retry needs to happen
-- we need to use GetContainerItemInfo to get a count, but it's not valid if an itemID isn't data cached
function stash.tracking:UpdateContainerSlot(container,bag,slot)
    local itemID = self:GetContainerItemID(bag,slot) -- this value is reliable even if not data cached
    if itemID then -- only filling slots that have an item
        if C_Item.IsItemDataCachedByID(itemID) then -- if item data cached, then its count is reliable
            local _,count,_,_,_,_,link = self:GetContainerItemInfo(bag,slot)
            if itemID==C.PET_CAGE_ITEMID then -- this is a caged pet which fortunately never stacks, save it as "battlepet:speciesID"
                container[bag][slot] = (link or ""):match("(battlepet:%d+)") or itemID
            elseif count and count>1 then
                container[bag][slot] = format("%d:%d",itemID,count)
            else
                container[bag][slot] = itemID
            end
            return true
        else -- itemID isn't data cached, which we need to get a count of items in the bag,slot; request data
            C_Item.RequestLoadItemDataByID(itemID)
            return false
        end
    elseif container[bag] then -- no itemID, empty it out
        container[bag][slot] = nil
        return true
    end
end

-- updates the bag,slot for the guild bank where tab = bag
function stash.tracking:UpdateGuildBankSlot(container,tab,slot)
    local _,count = GetGuildBankItemInfo(tab,slot)
    if count then
        local link = GetGuildBankItemLink(tab,slot)
        if link then
            local itemID = tonumber(link:match("item:(%d+)"))
            if itemID then
                if itemID==C.PET_CAGE_ITEMID then -- caged pet, pick up speciesID from the tooltip (link would be of Pet Cage)
                    StashGameTooltip:SetOwner(stash.frame,"ANCHOR_NONE")
                    local speciesID = StashGameTooltip:SetGuildBankItem(tab,slot)
                    if speciesID and speciesID>0 then
                        container[tab][slot] = format("battlepet:%d",speciesID)
                    end
                elseif count>1 then
                    container[tab][slot] = format("%d:%d",itemID,count)
                else
                    container[tab][slot] = itemID
                end
                return true
            else -- link exists but no item in it; it's not data cached
                return false
            end
        else
            return false -- link is there but incomplete; can't know what itemID this is yet
        end
    else
        container[tab][slot] = nil
        return true
    end
end

-- updates all currency for current character
local collapsedCurrencies = {} -- lookup table to backup collapsed currency headers
function stash.tracking:UpdateAllCurrency()
    -- backup collapsed headers
    for i=1,C_CurrencyInfo.GetCurrencyListSize() do
        local info = C_CurrencyInfo.GetCurrencyListInfo(i)
        if info.isHeader and not info.isHeaderExpanded then
            collapsedCurrencies[info.name] = true
        end
    end
    -- expand all headers
    for i=C_CurrencyInfo.GetCurrencyListSize(),1,-1 do
        local info = C_CurrencyInfo.GetCurrencyListInfo(i)
        if collapsedCurrencies[info.name] then
          C_CurrencyInfo.ExpandCurrencyList(i,true)
        end
    end
    -- gather currency
    wipe(stash.current.currency)
    for i=1,C_CurrencyInfo.GetCurrencyListSize() do
        local info = C_CurrencyInfo.GetCurrencyListInfo(i)
        if not info.isHeader and info.quantity and info.quantity>0 then
            local link = (C_CurrencyInfo.GetCurrencyListLink(i) or ""):match("(currency:%d+)")
            if link then
                stash.current.currency[link] = info.quantity
            end
        end
    end
    -- restore headers
    for i=C_CurrencyInfo.GetCurrencyListSize(),1,-1 do
        local info = C_CurrencyInfo.GetCurrencyListInfo(i)
        if collapsedCurrencies[info.name] then
            C_CurrencyInfo.ExpandCurrencyList(i,false)
        end
    end

    wipe(collapsedCurrencies)
end

-- mail

-- when mail is sent, if any items are involved then save them in pendingSendMail
-- when/if mail is successfully sent, these items will be added to recipient's stash if
-- they have a stash.saved node set up.
-- note the dot notation! this is a hook of a Blizzard API without a self reference
function stash.tracking.SendMailHook(recipient)
    wipe(pendingSendMail)
    -- add any items being sent to pendingSendMail with its count
    for i=1,ATTACHMENTS_MAX_SEND do
        if HasSendMailItem(i) then
            local _,itemID,_,count = GetSendMailItem(i)
            if itemID==C.PET_CAGE_ITEMID then -- item is a caged pet
                local link = GetSendMailItemLink(i)
                itemID = link:match("(battlepet:%d+)")
                if itemID then
                    tinsert(pendingSendMail,itemID)
                end
            elseif count>1 then
                tinsert(pendingSendMail,format("%d:%d",itemID,count))
            elseif itemID  then
                tinsert(pendingSendMail,itemID)
            end
        end
    end
    -- if an item is being sent, insert the realm,player the item(s) are being sent to
    if #pendingSendMail>0 then
        local recipientRealm = GetRealmName()
        if recipient:match("-") then -- send a BoA item to another realm
            recipient,recipientRealm = recipient:match("(.+)-(.+)")
        end
        tinsert(pendingSendMail,1,{recipientRealm,recipient})
    end
end

-- if mail was successfully sent, check if any pending items were being sent
function stash.tracking:MAIL_SEND_SUCCESS()
    if #pendingSendMail>1 then
        local realm,player = pendingSendMail[1][1],pendingSendMail[1][2]
        local otherStash = stash.saved[realm] and stash.saved[realm][player]
        if otherStash then
            for i=2,#pendingSendMail do
                tinsert(otherStash.mail,pendingSendMail[i])
            end
        end
    end
    wipe(pendingSendMail)
end

-- when mail closes, any items in SendMail are released
function stash.tracking:MAIL_CLOSED()
    wipe(pendingSendMail)
end

-- when opening mail, update the current player's mail
function stash.tracking:MAIL_INBOX_UPDATE()
    wipe(stash.current.mail) -- we're checking mail now, can drop stuff that was noted as sent
    for i=1,GetInboxNumItems() do
        if select(8,GetInboxHeaderInfo(i)) then -- if this inbox item has items
            for j=1,ATTACHMENTS_MAX_RECEIVE do
                if HasInboxItem(i,j) then
                    local _,itemID,_,count = GetInboxItem(i,j)
                    if itemID==C.PET_CAGE_ITEMID then -- item is a caged pet
                        -- GetInboxItemLink doesn't use a battlepet link :( so picking up speciesID from tooltip
                        StashGameTooltip:SetOwner(stash.frame,"ANCHOR_NONE")
                        local _,speciesID = StashGameTooltip:SetInboxItem(i,j)
                        if speciesID and speciesID>0 then
                            tinsert(stash.current.mail,format("battlepet:%d",speciesID))
                        end
                        StashGameTooltip:Hide()
                    elseif count>1 then
                        tinsert(stash.current.mail,format("%d:%d",itemID,count))
                    elseif itemID  then
                        tinsert(stash.current.mail,itemID)
                    end
                end
            end
        end
    end
end

local voidTimeout = 5
function stash.tracking:VOID_TRANSFER_DONE()
    if not CanUseVoidStorage() then
        return -- player doesn't have voide storage unlocked, leave
    end
    local found
    for tab=1,C.VOID_STORAGE_MAX_TABS do
        for slot=1,C.VOID_STORAGE_MAX_SLOTS do
            local itemID = GetVoidItemInfo(tab,slot)
            stash.current.void[tab][slot] = itemID
            if itemID then
                found = true
                voidTimeout = 0 -- only retry first attempt
            end
        end
    end
    if not found and voidTimeout>0 then
        voidTimeout = voidTimeout - 1
        C_Timer.After(1,stash.tracking.VOID_TRANSFER_DONE)
    end
end
